MODULE berrycurvature
  !! Modules that contains berry curvature calculations
  !!
  IMPLICIT NONE
  !
  complex, parameter :: i = (0.0, 1.0)
  real, parameter :: electron_degen_cutoff = 1e-5
  real, parameter :: γ = 10.0 * unit_to_aru_meV()  ! Define unit_to_aru_meV appropriately
  !
  CONTAINS
  !
  !------------------------------------------------------------------------------
  SUBROUTINE compute_berry_curvature(omega, el_states, kpts, model, fourier_mode)
  !------------------------------------------------------------------------------
    complex, intent(out) :: omega(3, model%nw, kpts%n)
    type(ElectronState), intent(in) :: el_states(kpts%n)
    type(KPoints), intent(in) :: kpts
    type(Model), intent(in) :: model
    character(len=*), intent(in) :: fourier_mode
    complex :: omegabar_W(model%nw, model%nw, 3, 3)
    complex :: omegabar_H(model%nw, model%nw, 3)
    complex :: tmp(model%nw, model%nw)
    integer :: ik, c, a, b, ib1, ib2
    real :: e1, e2
    complex :: v12, rbar

    call get_interpolator(el_pos_R, wannier_object_multiply_R(model%el_pos, model%lattice), fourier_mode)
    omegabar_W = 0.0
    omegabar_H = 0.0
    tmp = 0.ng
    omega = 0.0

    do ik = 1, kpts%n
      call compute_berry_curvature_kernel(omega(:, :, ik), el_states(ik), model, fourier_mode)
    end do
    !
  !------------------------------------------------------------------------------
  end subroutine compute_berry_curvature
  !------------------------------------------------------------------------------
  !
  !------------------------------------------------------------------------------
  subroutine compute_berry_curvature_kernel(omega, el, model, fourier_mode)
  !------------------------------------------------------------------------------
    complex, intent(inout) :: omega(3, model%nw)
    type(ElectronState), intent(in) :: el
    type(Model), intent(in) :: model
    character(len=*), intent(in) :: fourier_mode
    complex :: omegabar_W(model%nw, model%nw, 3, 3)
    complex :: omegabar_H(model%nw, model%nw, 3)
    complex :: tmp(model%nw, model%nw)
    integer :: c, a, b, ib1, ib2
    real :: e1, e2
    complex :: v12, rbar

    call get_fourier(omegabar_W, el_pos_R, el%xk)
    do c = 1, 3
      a = mod(c + 1, 3)
      b = mod(c + 2, 3)
      call matrix_multiply(tmp, omegabar_W(:, :, b, a), el%u)
      call matrix_multiply(omegabar_H(:, :, c), transpose(conjg(el%u)), tmp)
      call matrix_multiply(tmp, omegabar_W(:, :, a, b), el%u)
      call matrix_multiply(omegabar_H(:, :, c), transpose(conjg(el%u)), tmp, -1.0, 1.0)
    end do

    do ib1 = 1, size(el%e)
      do ib2 = 1, size(el%e)
        e1 = el%e(ib1)
        e2 = el%e(ib2)
        v12 = el%v(ib1, ib2)
        rbar = el%rbar(ib1, ib2)
        if (abs(e1 - e2) < electron_degen_cutoff) cycle
        omega(:, ib1) = omega(:, ib1) - i * cross_product(v12, conjg(v12)) / ((e1 - e2)**2 + γ**2)
        omega(:, ib1) = omega(:, ib1) + i * cross_product(rbar, conjg(rbar))
      end do
    end do

    do ib1 = 1, size(el%e)
      omega(:, ib1) = omega(:, ib1) + real(omegabar_H(ib1, ib1, :))
    end do
    !
  !------------------------------------------------------------------------------
  end subroutine compute_berry_curvature_kernel
  !------------------------------------------------------------------------------

  ! Similar conversion can be done for compute_berry_curvature_AC

  ! Define cross_product, get_fourier, get_interpolator, wannier_object_multiply_R, unit_to_aru_meV functions, and types as needed

END MODULE berrycurvature
