MODULE dirac_gap
  !! Modules that contains all the functions to calculate the energy gap at each k-point
  !!
  IMPLICIT NONE
  !
  CONTAINS
  !
  !------------------------------------------------------------!
  SUBROUTINE interpolatebands(k, nrr_k, irvec_r, energies)
  !------------------------------------------------------------!
  !
  USE kinds,            ONLY : DP
  USE wannier2bloch,    ONLY : hamwan2bloch
  USE global_var,       ONLY : chw
  USE input,            ONLY : use_ws, nbndsub
  USE cell_base,        ONLY : at, bg
  USE ep_constants,     ONLY : ryd2ev, czero, twopi, ci, zero
  !!!!
  IMPLICIT NONE
  !
  REAL(DP), INTENT(IN) :: k(3)
  !! K-point
  INTEGER, INTENT(IN) :: nrr_k
  !! Number of WS points for electrons
  REAL(DP), INTENT(IN) :: irvec_r(3, nrr_k)
  !! Wigner-Size supercell vectors, store in real instead of integer
  REAL(DP), INTENT(OUT)  :: energies(nbndsub)
  !! Energy eigenvalues
  INTEGER :: ierr
  !! Error status
  REAL(KIND = DP) :: rdotk(nrr_k)
  !! $r\cdot k$
  COMPLEX(KIND = DP) :: cuf(nbndsub, nbndsub)
  !! Rotation matrix, fine mesh, points k
  COMPLEX(KIND = DP), ALLOCATABLE :: cfac(:)
  !! Exponential factor cfac(nrr_k)
  !
  !
  ALLOCATE(cfac(nrr_k), STAT = ierr)
  IF (ierr /= 0) CALL errore('dirac_gap', 'Error allocating cfac', 1)
  !
  CALL DGEMV('t', 3, nrr_k, twopi, irvec_r, 3, k, 1, 0.0_DP, rdotk, 1)
  cfac(:)   = EXP(ci * rdotk(:))
  ! hamiltonian : Wannier -> Bloch
  ! Diagonalization of Hamiltonian
  CALL hamwan2bloch(nbndsub, nrr_k, cuf, energies, chw, cfac, .FALSE.)
  !
  DEALLOCATE(cfac, STAT = ierr)
  IF (ierr /= 0) CALL errore('dirac_gap', 'Error deallocating cfac', 1)
  !
  !------------------------------------------------------------!
  END SUBROUTINE
  !------------------------------------------------------------!
  !
  !------------------------------------------------------------!
  REAL(DP) FUNCTION find_energygap(k, nrr_k, irvec_r)
  !------------------------------------------------------------!
  !
  USE kinds,            ONLY : DP
  USE global_var,       ONLY : ivbm
  USE input,            ONLY : nbndsub
  !
  IMPLICIT NONE
  !
  REAL(DP), INTENT(IN) :: k(3)
  !! K-point
  INTEGER, INTENT(IN) :: nrr_k
  !! Number of WS points for electrons
  REAL(DP), INTENT(IN) :: irvec_r(3, nrr_k)
  !! Wigner-Size supercell vectors, store in real instead of integer
  REAL(DP) :: energy(nbndsub)
  !! Energy eigenvalues
  !
  CALL interpolatebands(k, nrr_k, irvec_r, energy)
  find_energygap = energy(ivbm + 1) - energy(ivbm)
  !
  !------------------------------------------------------------!
  END FUNCTION find_energygap
  !------------------------------------------------------------!
  !
  !------------------------------------------------------------!
  REAL(DP) FUNCTION find_surfacegap(k, nrr_k, irvec_r)
  !------------------------------------------------------------!
  ! Get the magnitude of the gap between the Fermi level and the nearest
  ! occupied band
  USE kinds,            ONLY : DP
  USE pwcom,            ONLY : ef
  USE input,            ONLY : nbndsub
  !!!!
  IMPLICIT NONE
  !
  REAL(DP), INTENT(IN) :: k(3)
  !! K-point
  INTEGER, INTENT(IN) :: nrr_k
  !! Number of WS points for electrons
  REAL(DP), INTENT (IN) :: irvec_r(3, nrr_k)
  !! Wigner-Size supercell vectors, store in real instead of integer
  REAL(DP) :: energies(nbndsub)
  !! Energy eigenvalues
  !
  CALL interpolatebands(k, nrr_k, irvec_r, energies)
  find_surfacegap = MINVAL(ABS(energies - ef))
  !------------------------------------------------------------!
  END FUNCTION
  !------------------------------------------------------------!
  !
END MODULE dirac_gap
