MODULE Dirac_Kpoints
  !! Modules that contains routines to calculate Dirac gap k-points
  !! 
  IMPLICIT NONE
  !
  CONTAINS
  !
  !------------------------------------------------------------------------------
  SUBROUTINE NelderMead(k0, length, func, minimum, gap, iter)
  !------------------------------------------------------------------------------
  ! this program solves the problem to find the minima of a function in a 
  ! direct search Nelder-Mead method
  ! This subroutine gives the local minimal of the func_gap in multidimensional
  ! parameters space
  ! Ref:  Nelder, J.A., and Mead, R. 1965, Computer Journal, vol. 7, pp. 308–313.
  !
  USE kinds,          ONLY : DP
  USE ep_constants,   ONLY : eps16
  USE io_global,      ONLY : stdout
  !
  !
  ! define constants
  INTEGER, PARAMETER :: n = 3     ! Dimensionality of the optimization
  INTEGER, PARAMETER :: m = n+1   !number of vertices in the simplex
  INTEGER, PARAMETER :: max_iter = 1000  !  number of iterations
  REAL(DP), PARAMETER :: alpha = 1.0, beta = 0.5, gamma = 2.0, delta = 0.5
  REAL(DP), PARAMETER :: thr = 1.d-6   ! Convergence threshold
  !
  ! define variables
  REAL(DP) :: gap_new, gap_new2
  !! Gaps of candidate points
  REAL(DP) :: simplex(n, m)
  REAL(DP) :: fval(m)
  REAL(DP) :: centroid(n)
  REAL(DP) :: reflected(n)
  REAL(DP) :: expanded(n)
  REAL(DP) :: contracted(n)
  INTEGER :: best_idx, worst_idx, worst2_idx
  INTEGER :: i, j, k
  ! function to calculate the energy gap for a given k point x
  !
  REAL(DP), INTENT(IN) :: k0(3)
  !! Starting guess for minimization
  REAL(DP), INTENT(IN) :: length
  !! Side length of the initial simplex
  REAL(DP), EXTERNAL :: func
  !! Function to minimize
  REAL(DP), INTENT(OUT) :: minimum(3)
  !! Final minimum
  REAL(DP), INTENT(OUT) :: gap
  !! Band gap at the minimum
  INTEGER, INTENT(OUT) :: iter
  !! Final number of iterations
  !
  ! Initialize the simplex
  simplex(:, 1) = k0
  simplex(:, 2) = k0 + (/ length, 0.00d0, 0.00d0 /)
  simplex(:, 3) = k0 + (/ 0.00d0, length, 0.00d0 /)
  simplex(:, 4) = k0 + (/ 0.00d0, 0.00d0, length /)
  !
  ! Evaluate objective function at the initial vertices
  DO i = 1, m
    fval(i) = func(simplex(:, i))
  END DO
  !
  ! Main optimization loop
  DO k = 1, max_iter
    ! Find the best, worst, and second worst vertices
    best_idx = 1 
    worst_idx = 1
    worst2_idx = 1
    DO i = 2, m 
      IF (fval(i) < fval(best_idx)) THEN
        best_idx = i
      ELSEIF (fval(i) > fval(worst_idx)) THEN
        worst_idx = i
      ELSEIF (fval(i) > fval(worst2_idx)) THEN
        worst2_idx = i
      ENDIF
    ENDDO
    IF (fval(best_idx) <= eps16 .OR. &
      (fval(worst_idx) - fval(best_idx))/fval(best_idx) < thr) &
      EXIT
    !
    ! Calculate centroid (excluding the worst point)
    centroid = 0.0
    DO i = 1, m 
      IF (i /= worst_idx) THEN
        centroid = centroid + simplex(:, i)
      ENDIF
    ENDDO
    centroid = centroid / REAL(n)
    !
    ! Reflect the worst point
    reflected = centroid + alpha * (centroid - simplex(:, worst_idx))
    gap_new = func(reflected)
    IF (gap_new < fval(best_idx)) THEN
      ! Candidate is the new best point
      expanded = centroid + gamma * (reflected - centroid)
      gap_new2 = func(expanded)
      IF (gap_new2 < fval(best_idx)) THEN
        ! Expand
        simplex(:, worst_idx) = expanded
        fval(worst_idx) = gap_new2
      ELSE
        simplex(:, worst_idx) = reflected
        fval(worst_idx) = gap_new
      ENDIF
    ELSE IF (gap_new < fval(worst2_idx)) THEN
      ! Candidate is not the worst point
      simplex(:, worst_idx) = reflected
      fval(worst_idx) = gap_new
    ELSE
      ! Candidate is worse than all points, except maybe the worst
      IF (gap_new < fval(worst_idx)) THEN
        ! Candidate is better than the old worst
        simplex(:, worst_idx) = reflected
        fval(worst_idx) = gap_new
      ENDIF
      contracted = centroid + beta* (simplex(:, worst_idx) - centroid)
      gap_new = func(contracted)
      IF (gap_new < fval(worst_idx)) THEN
        ! Contract
        simplex(:, worst_idx) = contracted
        fval(worst_idx) = gap_new
      ELSE
        ! Shrink
        DO i = 1, m 
          IF (i /= best_idx) THEN 
            simplex(:, i) = simplex(:, best_idx) + &
                    delta * (simplex(:, i) - simplex(:, best_idx))
            fval(i) = func(simplex(:, i))
          ENDIF
        ENDDO
      ENDIF
    ENDIF
    !
  ENDDO
  !
  IF (k == max_iter) WRITE(stdout, *) "WARNING:  Nelder-Mead did not converge"
  !
  minimum = simplex(:, best_idx)
  gap = fval(best_idx)
  iter = k
  !
  RETURN
  !  
  !--------------------------------------------------------------------------
  END SUBROUTINE NelderMead
  !--------------------------------------------------------------------------
  !
  !--------------------------------------------------------------------------
  SUBROUTINE find_irreducible_minima(func, threshold, dist_threshold, &
                                     nk1, nk2, nk3, count, points)
  !--------------------------------------------------------------------------
  !! Find all distinct, irreducible local minima of the given function over the
  !! Brillouin zone.
  USE kinds,          ONLY : DP
  !
  IMPLICIT NONE
  !
  REAL(DP), EXTERNAL :: func
  !! Function whose minima will be found
  REAL(DP), INTENT(IN) :: threshold
  !! Maximum value of a point to be kept
  REAL(DP), INTENT(IN) :: dist_threshold
  !! Minimum separation to be considered a distinct point
  INTEGER, INTENT(IN) :: nk1, nk2, nk3
  !! Initial guess mesh
  INTEGER, INTENT(OUT) :: count
  !! Number of distinct points found
  REAL(DP), ALLOCATABLE, INTENT(OUT) :: points(:, :)
  !! Locations of points
  !
  INTEGER :: ierr
  !! Status of memory allocation
  INTEGER :: iter
  !! Final number of iterations
  INTEGER :: i, j, k, l
  !! Loop iterators
  REAL(DP) :: gap
  !! Value at the minimum
  REAL(DP) :: distance
  !! Distance between points
  REAL(DP) :: length
  !! Side length of the initial simplex for Nelder-Mead
  REAL(DP) :: minimum(3)
  !! Final minimum
  REAL(DP), ALLOCATABLE :: klist(:, :)
  !! Temporary list of points
  !
  ! We must be able to accommodate every starting point finding a unique minimum
  ALLOCATE(klist(3, nk1*nk2*nk3), STAT = ierr)
  IF (ierr /= 0) CALL errore('find_irreducible_minima', &
                  'Error allocating points', 1)
  count = 0
  length = 1.d0 / MAX(nk1, nk2, nk3)
  !
  DO i = 0, nk1 - 1
    DO j = 0, nk2 - 1
      point:DO k = 0, nk3 - 1
        CALL NelderMead((/DBLE(i)/nk1, DBLE(j)/nk2, DBLE(k)/nk3/), &
                        length, func, minimum, gap, iter)
        !
        ! We discard if the value is higher than threshold
        IF (gap > threshold) CYCLE point
        !
        CALL reduce(minimum)
        !
        DO l = 1, count
          distance = SQRT((minimum(1)-klist(1, l))**2 &
                          +(minimum(2)-klist(2, l))**2 &
                          +(minimum(3)-klist(3, l))**2)
          IF (distance < dist_threshold) CYCLE point
        ENDDO
        !
        count = count + 1
        klist(:, count) = minimum
      ENDDO point
    ENDDO
  ENDDO
  !
  ALLOCATE(points(3, count), STAT = ierr)
  IF (ierr /= 0) CALL errore('find_irreducible_minima', &
                  'Error allocating points', 1)
  !
  points(:, 1:count) = klist(:, 1:count)
  !
  DEALLOCATE(klist, STAT = ierr)
  IF (ierr /= 0) CALL errore('find_irreducible_minima', &
                  'Error deallocating points', 1)
  END SUBROUTINE
  !
  !--------------------------------------------------------------------------
  SUBROUTINE find_topological_points(nrr_k, irvec_r, count, points)
  !--------------------------------------------------------------------------
  !! Find all irreducible topological points
  !
  USE ep_constants,     ONLY : ryd2mev 
  USE input,            ONLY : nkc1, nkc2, nkc3
  USE kinds,            ONLY : DP
  !
  IMPLICIT NONE
  !
  REAL(DP), PARAMETER :: gap_threshold = 1.d0 / ryd2mev
  !! Maximum band gap to be considered closed
  REAL(DP), PARAMETER :: dist_threshold = 0.001
  !! Minimum separation to be considered a distinct point
  !
  INTEGER, INTENT(IN) :: nrr_k
  REAL(DP), INTENT(IN) :: irvec_r(3, nrr_k)
  INTEGER, INTENT(OUT) :: count
  !! Number of distinct topological points found
  REAL(DP), ALLOCATABLE, INTENT(OUT) :: points(:, :)
  !! Locations of topological points
  !
  CALL find_irreducible_minima(gap_func, gap_threshold, dist_threshold, &
                               nkc1, nkc2, nkc3, count, points)
  !
  CONTAINS
  ! Wrapper function to minimize
  REAL(DP) FUNCTION gap_func(x)
    USE kinds,          ONLY : DP
    USE dirac_gap,      ONLY : find_energygap
    !
    IMPLICIT NONE
    !
    REAL(DP) :: x(3)
    !
    gap_func = find_energygap(x, nrr_k, irvec_r)
  END FUNCTION
  !--------------------------------------------------------------------------
  END SUBROUTINE
  !--------------------------------------------------------------------------
  !
  !--------------------------------------------------------------------------
  SUBROUTINE find_fermi_surface(nrr_k, irvec_r, count, points)
  !--------------------------------------------------------------------------
  !! Find points along the Fermi surface
  USE input,           ONLY : fsthick, nkf1, nkf2, nkf3
  USE kinds,           ONLY : DP
  !
  IMPLICIT NONE
  !
  REAL(DP), PARAMETER :: dist_threshold = 0.01
  !! Minimum separation to be considered a distinct point
  !
  INTEGER, INTENT(IN) :: nrr_k
  REAL(DP), INTENT (IN) :: irvec_r(3, nrr_k)
  INTEGER, INTENT(OUT) :: count
  !! Number of distinct topological points found
  REAL(DP), ALLOCATABLE, INTENT(OUT) :: points(:, :)
  !! Locations of topological points
  !
  CALL find_irreducible_minima(surface_func, fsthick, dist_threshold, &
                               nkf1, nkf2, nkf3, count, points)
  !
  CONTAINS
  ! Wrapper function to minimize
  REAL(DP) FUNCTION surface_func(x)
    USE kinds,          ONLY : DP
    USE dirac_gap,      ONLY : find_surfacegap
    !
    IMPLICIT NONE
    !
    REAL(DP) :: x(3)
    !
    surface_func = find_surfacegap(x, nrr_k, irvec_r)
  END FUNCTION
  !--------------------------------------------------------------------------
  END SUBROUTINE
  !--------------------------------------------------------------------------
  !
  !--------------------------------------------------------------------------
  SUBROUTINE reduce(x)
  !--------------------------------------------------------------------------
  !! Map a value to its irreducible equivalent in the first octant of the
  !! first Brillouin zone.  The point returned will have the smallest
  !! positive coordinates of all equivalent points.
  USE kinds,          ONLY : DP
  USE symm_base,      ONLY : s, t_rev, nsym
  !
  IMPLICIT NONE
  !
  REAL(DP), INTENT(INOUT) :: x(3)
  !! Point to reduce
  INTEGER :: i, ns
  !! Iteration variables
  REAL(DP) :: best(3)
  !! Best transformed point
  REAL(DP) :: current(3)
  !! Current candidate
  !
  best = 10.d0
  DO ns = 1, nsym
    DO i = 1, 3
      current(i) = s(i, 1, ns) * x(1) &
          + s(i, 2, ns) * x(2) &
          + s(i, 3, ns) * x(3)
      current(i) = current(i) - NINT(current(i))
    ENDDO
    IF(t_rev(ns) == 1) current = -current
    current = current - FLOOR(current)
    IF (current(1) < best(1) &
      .AND. current(2) < best(2) &
      .AND. current(3) < best(3)) best = current
  ENDDO
  x = best
  !--------------------------------------------------------------------------
  END SUBROUTINE reduce
  !--------------------------------------------------------------------------
  !
!--------------------------------------------------------------------------
END MODULE Dirac_Kpoints
!--------------------------------------------------------------------------
